package ru.dbotthepony.mc.otm.client.render

import com.google.gson.JsonDeserializationContext
import com.google.gson.JsonDeserializer
import com.google.gson.JsonElement
import com.google.gson.JsonPrimitive
import com.google.gson.JsonSerializationContext
import com.google.gson.JsonSerializer
import com.google.gson.TypeAdapter
import com.google.gson.internal.bind.TypeAdapters
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonWriter
import net.minecraft.client.renderer.texture.TextureAtlasSprite
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.resources.ResourceLocation
import ru.dbotthepony.mc.otm.isClient
import java.lang.reflect.Type
import java.util.stream.Stream

fun AtlasSkinElement(location: ResourceLocation) = AtlasSkinElement.create(location)

class AtlasSkinElement private constructor(
	val location: ResourceLocation,
	marker: Any?
) : AbstractSkinElement() {
	private var changeset = -1
	private var _textureAtlasSprite: TextureAtlasSprite? = null

	private val textureAtlasSpriteImpl: TextureAtlasSprite get(): TextureAtlasSprite {
		check(isClient) { "Invalid realm" }
		val _textureAtlasSprite = _textureAtlasSprite

		if (_textureAtlasSprite == null || changeset != WidgetAtlasHolder.INSTANCE.changeset) {
			val get = WidgetAtlasHolder.INSTANCE.getSprite(location)
			this._textureAtlasSprite = get
			changeset = WidgetAtlasHolder.INSTANCE.changeset
			return get
		}

		return _textureAtlasSprite
	}

	val textureAtlasSprite: TextureAtlasSprite get(): TextureAtlasSprite {
		check(isWidgetAtlasAvailable) { "Atlas is not available; either we are not a client, or we are running datagen" }
		return textureAtlasSpriteImpl
	}

	override val x get() = textureAtlasSprite.x.toFloat()
	override val y get() = textureAtlasSprite.y.toFloat()
	override val width get() = textureAtlasSprite.width.toFloat()
	override val height get() = textureAtlasSprite.height.toFloat()

	override val u0 get() = textureAtlasSprite.u0
	override val v0 get() = textureAtlasSprite.v0
	override val u1 get() = textureAtlasSprite.u1
	override val v1 get() = textureAtlasSprite.v1
	override val texture: ResourceLocation get() = textureAtlasSprite.atlas().location()

	override fun equals(other: Any?): Boolean {
		if (other is AtlasSkinElement)
			return location == other.location

		return super.equals(other)
	}

	override fun hashCode(): Int {
		return WidgetAtlasHolder.INSTANCE.hashCode() xor location.hashCode()
	}

	override fun toString(): String {
		return "AtlasSkinElement[$location]"
	}

	override fun toJson(): JsonElement {
		return JsonPrimitive(location.toString())
	}

	override fun toNetwork(buff: FriendlyByteBuf) {
		buff.writeResourceLocation(location)
	}

	companion object : TypeAdapter<AtlasSkinElement>(), JsonSerializer<AtlasSkinElement>, JsonDeserializer<AtlasSkinElement> {
		private val skinElementCache = HashMap<ResourceLocation, AtlasSkinElement>()

		val keysStream: Stream<ResourceLocation> get() = skinElementCache.keys.stream()

		private fun queueRebuild() {
			WidgetAtlasHolder.INSTANCE.queueRebuild()
		}

		private fun createImpl(location: ResourceLocation): AtlasSkinElement {
			var element = skinElementCache[location]

			if (element == null) {
				element = AtlasSkinElement(location, null)
				skinElementCache[location] = element

				if (isWidgetAtlasAvailable) {
					queueRebuild()
				}
			}

			return element
		}

		@JvmStatic
		fun create(location: ResourceLocation): AtlasSkinElement {
			if (!isWidgetAtlasAvailable) {
				return AtlasSkinElement(location, null)
			}

			return createImpl(location)
		}

		fun fromNetwork(buff: FriendlyByteBuf): AtlasSkinElement {
			return create(buff.readResourceLocation())
		}

		fun fromJson(element: JsonElement): AtlasSkinElement {
			return create(ResourceLocation.tryParse(element.asString) ?: ResourceLocation("missing"))
		}

		override fun write(out: JsonWriter, value: AtlasSkinElement) {
			TypeAdapters.JSON_ELEMENT.write(out, value.toJson())
		}

		override fun read(`in`: JsonReader): AtlasSkinElement {
			return fromJson(TypeAdapters.JSON_ELEMENT.read(`in`))
		}

		override fun serialize(
			src: AtlasSkinElement,
			typeOfSrc: Type,
			context: JsonSerializationContext
		): JsonElement {
			return src.toJson()
		}

		override fun deserialize(
			json: JsonElement,
			typeOfT: Type,
			context: JsonDeserializationContext
		): AtlasSkinElement {
			return fromJson(json)
		}
	}
}
